"""Tests suite for ORBIT."""

__author__ = ["Alicia Key", "Parangat Bhaskar", "Annika Eberle", "Owen Roberts", "Garrett Barter"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Alicia Key"
__email__ = "alicia.key@nrel.gov"